/*
	Written by: FIREking
*/
//Have a look in demo.dm to see how to use /pixel_projectile!

//////////////////////////////////////
//				Defines				//
//////////////////////////////////////

#define WORLD_ICON_SIZE 64
#define WORLD_HALF_ICON_SIZE 32
#define WORLD_ICON_MULTIPLY 0.015625 //This is  1 / WORLD_ICON_SIZE
#define WORLD_MAX_PX 4160 //This is the size of your map in tiles + 1 times world_icon_size
#define WORLD_MAX_PY 4160 //This is the size of your map in tiles + 1 times world_icon_size
#define BOTS_SHOOT_PROJECTILES

//////////////////////////////////////
//			Debug Stuff				//
//////////////////////////////////////

#ifdef DEBUG
var/tmp/projectiles = 0
var/tmp/projectiles_fired = 0
#endif

//////////////////////////////////////
//			Utility Stuff			//
//////////////////////////////////////

proc/get_dir_adv(atom/ref, atom/target)
	//Written by Lummox JR
	//Returns the direction between two atoms more accurately than get_dir()

    if(target.z > ref.z) return UP
    if(target.z < ref.z) return DOWN

    . = get_dir(ref, target)
    if(. & . - 1)        // diagonal
        var/ax = abs(ref.x - target.x)
        var/ay = abs(ref.y - target.y)
        if(ax >= (ay << 1))      return . & (EAST | WEST)   // keep east/west (4 and 8)
        else if(ay >= (ax << 1)) return . & (NORTH | SOUTH) // keep north/south (1 and 2)
    return .

atom/movable/proc/set_pos_px(px, py)
	//this sets the atom's x, y and the atom's pixel_x, pixel_y from absolute pixel coordinates
	src.x = px * WORLD_ICON_MULTIPLY
	src.y = py * WORLD_ICON_MULTIPLY
	src.pixel_x = px % WORLD_ICON_SIZE - WORLD_HALF_ICON_SIZE
	src.pixel_y = py % WORLD_ICON_SIZE - WORLD_HALF_ICON_SIZE

//////////////////////////////////////
//			Target Stuff			//
//////////////////////////////////////

atom/var/tmp/atom/target = null //all atom's can have a single atom target
atom/var/tmp/list/targeted = list() //an atom can be targeted by several atoms

atom/proc/add_target(atom/a)
	//sets atom/a to src's target
	//adds src to atom/a's targeted list
	//not meant to be called directly

	if(!src.targeted) src.targeted = list()
	src.targeted += a

atom/proc/remove_target(atom/a)
	//removes atom/a from src's targeted list
	//removes src as atom/a's target
	//not meant to be called directly

	if(!src.targeted) return
	src.targeted -= a
	a.target = null
	if(!src.targeted.len) src.targeted = null

atom/proc/remove_targets()
	//removes all targets from targeted list
	//removes src from each target's target in targeted list
	//call this when you want to clear all targets

	if(!src.targeted) return
	for(var/atom/v in src.targeted)
		v.target = null
	src.targeted = null

atom/proc/set_target(atom/a)
	//sets src.target to atom/a
	//adds src to a's src.targeted list
	//call this when you want to acquire a target for src

	if(src.target)
		src.drop_target()
	src.target = a
	src.target.add_target(src)

atom/proc/drop_target()
	//removes current target
	//removes src from current target's targeted list
	//call this when you want to drop src's current target

	if(!src.target) return
	src.target.remove_target(src)
	src.target = null

//////////////////////////////////////
//			pixel_projectile		//
//////////////////////////////////////

pixel_projectile
	parent_type = /obj
	animate_movement = NO_STEPS
	var/tmp
		angle = 0
		px = 0 //the true pixel location of the projectile
		py = 0
		last_x = 0 //the last x/y coordinate of the tile we were in
		last_y = 0
		velocity = 12 //speed of the projectile
		atom/owner

		vx = 0 //vector px (movement increment)
		vy = 0 //vector py (movement increment)

	New(loc, vel = 0)

		#ifdef DEBUG
		projectiles++
		projectiles_fired++
		#endif

		if(!usr) del src
		src.owner = usr
		if(!src.owner.target) del src
		if(!src.owner.target.loc) del src
		if(src.owner.target == src.loc) del src

		src.loc = src.owner.loc

		if(vel) src.velocity = vel

		//ok now the real important stuff, initial position and vector calculation

		//set initial position
		src.px = src.x * WORLD_ICON_SIZE + WORLD_HALF_ICON_SIZE
		src.py = src.y * WORLD_ICON_SIZE + WORLD_HALF_ICON_SIZE

		if(!src.check_bounds()) del src

		var/tmp
			dx = 0 //delta x
			dy = 0 //delta y
			dr = 0 //delta root

		//delta x,y is absolute pixel x minus the starting absolute pixel x
		dx = (src.owner.target.x * WORLD_ICON_SIZE) + WORLD_HALF_ICON_SIZE - src.px
		dy = (src.owner.target.y * WORLD_ICON_SIZE) + WORLD_HALF_ICON_SIZE - src.py
		//root
		dr = sqrt(dx * dx + dy * dy)
		//vector x (amount to move x,y each frame)
		src.vx = dx / dr * src.velocity
		src.vy = dy / dr * src.velocity

		src.dir = get_dir_adv(src, src.owner.target)

		//that's it

		//begin!
		spawn src.update()

	Del()
		//garbage collector stuff
		if(loc)
			if(src in loc:collidables)
				loc:collidables -= src
			loc = null
			if(target)
				target.targeted -= src
				target = null
			owner = null
			return
		#ifdef DEBUG
		projectiles--
		#endif

		..()

	Cross(atom/movable/a)
		//if something happens to cross this tile while we're sitting in it before moving to next tile
		//we should check for collide!
		if(src.can_collide(a)) src.collide(a)
		return ..()

	proc/update()
		//update position
		if(!src.loc) return

		src.px += src.vx
		src.py += src.vy

		if(!src.check_bounds()) del src

		//update x, y, pixel_x, pixel_y to reflect new location
		src.set_pos_px(src.px, src.py)

		//did we enter a new tile?
		if(src.last_x != src.x || src.last_y != src.y)
			src.last_x = src.x
			src.last_y = src.y

			if(src.loc) loc:collide_here(src)
			//using colon operator assuming the projectile's loc will always be a turf

		spawn(world.tick_lag) src.update()

	proc/check_bounds()
		//checks to see if a pixel is in bounds of the map
		//if(src.px > WORLD_ICON_SIZE && src.py > WORLD_ICON_SIZE && src.px <= WORLD_MAX_PX && src.py <= WORLD_MAX_PY)
		if((src.px in WORLD_ICON_SIZE to WORLD_MAX_PX) && (src.py in WORLD_ICON_SIZE to WORLD_MAX_PY))
			return 1
		return 0

	proc/collide(atom/a)
		spawn(get_speed_delay(velocity) - 1) del src

	proc/can_collide(atom/a)
		if(isturf(a))
			if(a.density)
				return 1
		else
			if(a != src && a != src.owner)
				return 1
		return 0

//the collidables list allows us to avoid having to generate lists per tile update per projectile
//just make sure to remove your atom from this list at garbage collection!
turf/var/tmp/list/collidables = null
turf/Entered(atom/a)
	if(!collidables) collidables = list()
	collidables += a
	..()

turf/Exited(atom/a)
	if(!collidables) return
	collidables -= a
	if(!collidables.len) collidables = null
	..()

turf/proc/collide_here(pixel_projectile/p)
	if(p.can_collide(src))
		p.collide(src)
	if(collidables)
		for(var/atom/a in collidables)
			if(p.can_collide(a))
				p.collide(a)